#Multi Qubit Correction (MQC) for Quantum Annealers
#Ramin Ayanzadeh, John Dorband, Milton Halem and Tim Finin 
#Scientific Reports (In Press)
#August, 2021
#AR14: https://arxiv.org/abs/2010.00115

import networkx as nx
import copy

def reduce(A, B, h, J):
    '''
        Reduce two samples A and b to a new sample z with lower Ising energy value
        Arges:
            A--int:{+1/-1} dict: a sample, attained by a sampler
            B--int:{+1/-1} dict: a sample, attained by a sampler
            h-- int:float/int dict: linear coefficients of an Ising model
            J-- (int,int):float/int dict: quadratic coefficients of an Ising model
        Output:  
            z--int:{+1/-1} dict: resulting sample 
    '''                
    S=[]
    D=[]                
    for i in A.keys():
        if A[i] == B[i]:
            S.append(i)
        else:
            D.append(i)
    g=nx.Graph()        
    g.add_nodes_from(A.keys())        
    g.add_edges_from([e for e in J.keys() if J[e] !=0])
    g.remove_nodes_from(S)
    T=[list(c) for c in nx.connected_components(g)]                            
    z=dict(copy.deepcopy(A))
    for k in range(len(T)):            
        I_k=    0            
        for i in T[k]:                
            I_k += A[i]*h.get(i, 0.0)
            for j in S:
                J_ij=J.get((i,j), J.get((j,i), None))
                if J_ij:
                    I_k += A[i]*A[j]*J_ij                                
        if I_k >0:                
            for l in T[k]:
                z[l] *= -1        
    return z


def MQC(samples, h, J):
    '''
        Multi-Qubit Correction (MQC) is a postprocessing method for improving the reliability of quantum annealers
        Args:
            samples-- iterable (list/tuple) of samples int:{+1/-1 dict)
            h-- int:float/int dict: linear coefficients of an Ising model
            J-- (int,int):float/int dict: quadratic coefficients of an Ising model
        Output:
            z-- int:{+1/-1} dict: a solution for the Ising model
    '''                
    temp_pop=copy.deepcopy(samples)        
    while len(temp_pop) >1:            
        if len(temp_pop) %2 !=0:                        
            temp_pop.append(copy.deepcopy(temp_pop[0]))                
        next_pop=[]
        for i in range(round(len(temp_pop)/2)):                            
            A=temp_pop[i*2]
            B=temp_pop[(i*2)+1]                                
            next_pop.append(reduce(A, B, h, J))
        temp_pop=next_pop        
    return temp_pop[0]

